# ===============================================
# (C) 2018 AO Kaspersky Lab. All Rights Reserved.
# ===============================================

import binascii
import iu_config as config

DbType = config.APT_DB_TYPE


class SqlOpt:
  StrBin  = 0x01
  HashBin = 0x02


class SqlType:
  CHAR  = 'char'
  NCHAR = 'nchar'
  INT   = 'int'
  LONG  = 'long'
  BIT   = 'bit'
  TIME  = 'time'
  HASH  = 'hash'


class SqlColumn(object):
  def __init__(self, name, dataType, dataSize=None, isNull=False):
    self.name = name
    self.dataType = dataType
    self.dataSize = dataSize
    self.isNull = isNull

class SqlTable(object):
  def __init__(self,tableName,columns):
    self.tableName = tableName
    self.columns = columns


if (config.APT_DB_TYPE == 'MSSQL'):
  import pyodbc

  DbModule = pyodbc

  def ConnectDB():
    connstr = 'Driver={SQL Server};Integrated Security=true;Server=%s' % config.APT_DB_SERVER
    conn = pyodbc.connect(connstr)
    if config.APT_DB_NAME != '':
      with conn.cursor() as cur:
        cur.execute('use ['+config.APT_DB_NAME+']')
    return conn

  def Binary(s):
    return pyodbc.Binary(s)

  def Boolean(s):
    return s

  def MakeParmRef(idx):
    return '?'

  def QuoteName(s):
    qp = []
    for p in s.split('.'):
      qp.append('[%s]' % (p,))
    return '.'.join(qp)

  def ExecSP(cursor, name, args):
    suffix = ','.join(['?']*len(args))
    sql = 'exec %s %s' % (QuoteName(name), suffix)
    cursor.execute(sql, args)

  def ColumnSpec(col):
    typeSpec = None
    if col.dataType == SqlType.CHAR:
      if col.dataSize == None:
        typeSpec = 'varchar(max)'
      else:
        typeSpec = 'varchar(%s)' % (col.dataSize,)
    elif col.dataType == SqlType.NCHAR:
      if col.dataSize == None:
        typeSpec = 'nvarchar(max)'
      else:
        typeSpec = 'nvarchar(%d)' % (col.dataSize,)
    elif col.dataType == SqlType.INT:
      typeSpec = 'int'
    elif col.dataType == SqlType.LONG:
      typeSpec = 'bigint'
    elif col.dataType == SqlType.BIT:
      typeSpec = 'bit'
    elif col.dataType == SqlType.TIME:
      typeSpec = 'datetime'
    elif col.dataType == SqlType.HASH:
      typeSpec = 'binary(%d)' % (col.dataSize,)

    nullSpec = ''
    if col.isNull == True:
      nullSpec = 'null'
    elif col.isNull == False:
      nullSpec = 'not null'

    return '[%s] %s %s' % (col.name, typeSpec, nullSpec)

  def DoCreateTable(dbConn,tableSpec):
    query = 'create table %s (' % (QuoteName(tableSpec.tableName),)
    columnSpecs = []
    for col in tableSpec.columns:
      columnSpecs.append(ColumnSpec(col))
    query += ','.join(columnSpecs)
    query += ')'
    dbConn.ExecSql(query)

  UseSchemas = True
  UseBinaryHash = True
  CommitQuery = False
  UseInetAddr = True
  NativeMultiResults = True

elif (config.APT_DB_TYPE == 'PGSQL'):
  import psycopg2

  DbModule = psycopg2

  def ConnectDB():
    connParms = { 'host' : config.APT_DB_SERVER }
    if config.APT_DB_PORT != '':
      connParms['port'] = int(config.APT_DB_PORT)
    if config.APT_DB_USER != '':
      connParms['user'] = config.APT_DB_USER
    if config.APT_DB_PASSWORD != '':
      connParms['password'] = config.APT_DB_PASSWORD
    if (config.APT_DB_NAME != ''):
      connParms['database'] = config.APT_DB_NAME
    return psycopg2.connect(**connParms)

  def ExecSP(cursor, name, args):
    cursor.callproc(QuoteName(name), args)

  def Binary(s):
    return psycopg2.Binary(s)

  def Boolean(s):
    return bool(s)

  def MakeParmRef(idx):
    return '%s'

  def QuoteName(s):
    return '%s' % (s,)

  def ColumnSpec(col):
    typeSpec = None
    if col.dataType == SqlType.CHAR or col.dataType == SqlType.NCHAR:
      if col.dataSize == None:
        typeSpec = 'text'
      else:
        typeSpec = 'varchar(%s)' % (col.dataSize,)
    elif col.dataType == SqlType.INT:
      typeSpec = 'int'
    elif col.dataType == SqlType.LONG:
      typeSpec = 'bigint'
    elif col.dataType == SqlType.BIT:
      typeSpec = 'bit'
    elif col.dataType == SqlType.TIME:
      typeSpec = 'timestamp(3)'
    elif col.dataType == SqlType.HASH:
      typeSpec = 'text'

    nullSpec = ''
    if col.isNull == True:
      nullSpec = 'null'
    elif col.isNull == False:
      nullSpec = 'not null'

    return '%s %s %s' % (QuoteName(col.name), typeSpec, nullSpec)

  def DoCreateTable(dbConn,tableSpec):
    query = 'create table %s (' % (QuoteName(tableSpec.tableName),)
    columnSpecs = []
    for col in tableSpec.columns:
      columnSpecs.append(ColumnSpec(col))
    query += ','.join(columnSpecs)
    query += ')'
    dbConn.ExecSql(query)

  UseSchemas = True
  UseBinaryHash = False
  CommitQuery = True
  UseInetAddr = True
  NativeMultiResults = False

elif (config.APT_DB_TYPE == ''):

  def ConnectDB():
    raise Exception('DB interface not implemented')

  def ExecSP(cursor, name, args):
    raise Exception('DB interface not implemented')

  def Binary(s):
    return s

  def Boolean(s):
    return s

  UseSchemas = True
  UseBinaryHash = False
  CommitQuery = False
  UseInetAddr = True
  NativeMultiResults = False

else:
  raise Exception('Unknown APT_DB_TYPE: %s' % (config.APT_DB_TYPE,))


def Hex2Bin(s):
  pfx = s[0:2]
  if pfx=='0x' or pfx=='0X':
    s = s[2:]
  return binascii.a2b_hex(s)

class AptDbConnection:
  def __init__(self):
    self.connection = ConnectDB()
    self.needCheck = False
    self.closed = False

  def QueryProc(self,proc_name,args):
    if self.closed:
      raise IOError('QueryProc operation on closed AptDbConnection')
    try:
      cursor = self.CheckConnection().cursor()
      ExecSP(cursor, proc_name, args)
      if CommitQuery:
        self.connection.commit()
      return cursor
    except Exception:
      self.needCheck = True
      self.connection.rollback()
      raise

  def ExecProc(self,proc_name,args):
    if self.closed:
      raise IOError('ExecProc operation on closed AptDbConnection')
    try:
      with self.CheckConnection().cursor() as cursor:
        ExecSP(cursor, proc_name,args)
        self.connection.commit()
    except Exception:
      self.needCheck = True
      self.connection.rollback()
      raise

  def QuerySql(self,sql,parms=()):
    if self.closed:
      raise IOError('QuerySql operation on closed AptDbConnection')
    try:
      cursor = self.CheckConnection().cursor()
      cursor.execute(sql,parms)
      if CommitQuery:
        self.connection.commit()
      return cursor
    except Exception:
      self.needCheck = True
      self.connection.rollback()
      raise

  def ExecSql(self,sql,parms=()):
    if self.closed:
      raise IOError('ExecSql operation on closed AptDbConnection')
    try:
      with self.CheckConnection().cursor() as cursor:
        cursor.execute(sql,parms)
        self.connection.commit()
    except Exception:
      self.needCheck = True
      self.connection.rollback()
      raise

  def InsertRecord(self, tableName, parms):
    if self.closed:
      raise IOError('InsertRecord operation on closed AptDbConnection')
    nameList = []
    parmList = []
    valList = []
    parmIdx = 0
    for parm in parms:
      nameList.append(parm[0])
      parmList.append(MakeParmRef(parmIdx))
      val = parm[1]
      if len(parm) > 2:
        opt = parm[2]
        if (opt & SqlOpt.StrBin) != 0:
          val = Binary(Hex2Bin(val))
        elif (opt & SqlOpt.HashBin) != 0:
          if UseBinaryHash:
            val = Binary(Hex2Bin(val))
      valList.append(val)
      parmIdx += 1

    fieldsSpec = ','.join(nameList)
    parmsSpec = ','.join(parmList)
    vals = tuple(valList)

    sql = 'INSERT INTO %s (%s) VALUES (%s)' % (QuoteName(tableName),fieldsSpec,parmsSpec)
    self.ExecSql(sql, vals)


  def FetchHeaders(self,cursor):
    headers = []
    for desc in cursor.description:
      headers.append(str(desc[0]))
    return headers

  def FetchRows(self,cursor):
    rows = []
    for row in cursor:
      rows.append(list(row))
    return rows

  def QueryMultiProc(self, proc_name, args):
    if self.closed:
      raise IOError('QueryMultiProc operation on closed AptDbConnection')
    headerSet = []
    dataSet = []
    if NativeMultiResults:
      try:
        with self.QueryProc(proc_name, args) as cursor:
          while True:
            headers = self.FetchHeaders(cursor)
            rows = self.FetchRows(cursor)
            headerSet.append(headers)
            dataSet.append(rows)
            if not cursor.nextset():
              break
      except Exception:
        self.connection.rollback()
        raise
    else:
      try:
        resultTables = []
        with self.QueryProc(proc_name, args) as cursor:
          for r in cursor:
            resultTables.append(str(r[0]))

        for tableName in resultTables:
          query = 'select * from %s' % (QuoteName(tableName),)
          with self.QuerySql(query,()) as cursor:
            headers = self.FetchHeaders(cursor)
            rows = self.FetchRows(cursor)
            headerSet.append(headers)
            dataSet.append(rows)
      except Exception:
        self.connection.rollback()
        raise

    return dataSet,headerSet

  def CreateTable(self,tableSpec):
    if self.closed:
      raise IOError('CreateTable operation on closed AptDbConnection')
    DoCreateTable(self,tableSpec)

  def DropTable(self,tableName):
    if self.closed:
      raise IOError('DropTable operation on closed AptDbConnection')
    self.ExecSql('drop table %s' % (QuoteName(tableName),))

  def InsertTableRecord(self,tableSpec,values):
    if self.closed:
      raise IOError('InsertTableRecord operation on closed AptDbConnection')
    insParms = []
    columns = tableSpec.columns
    if len(columns) != len(values):
      raise ValueError('Number of values does not match number of columns')
    c = 0
    while c < len(columns):
      col = columns[c]
      opt = 0
      if col.dataType == SqlType.HASH:
        opt |= SqlOpt.HashBin
      insParms.append((col.name,values[c],opt))
      c += 1
    self.InsertRecord(None,tableSpec.tableName,insParms)

  def CheckConnection(self):
    if self.needCheck == False:
      return self.connection
    try:
      with self.connection.cursor() as cursor:
        cursor.execute('select 1')
        self.needCheck = False
        return self.connection
    except Exception:
      pass
    try:
      self.connection.Close()
    except Exception:
      pass
    try:
      self.connection = ConnectDB()
    except Exception:
      pass
    return self.connection

  def Close(self):
    if not self.closed:
      self.connection.close()
      self.closed = True

  def __enter__(self):
    return self

  def __exit__(self, *args):
    self.Close()
